#ifndef AutoGatherAI_H
#define AutoGatherAI_H

#include "GameChanges.H"
#include "ClientModule.H"
#include "EventHandler.H"
#include "Game.H"

#include "GobPtrWrapper_G3.H"
#include "TerrainModule.H"
#include "SimpleTerrain.H"

#include <set>
#include <map>
#include <string>
#include "boost/shared_ptr.hpp"


/// Automatically manage the resource gathering operations, assuming a single control center.
/** This is an implementation for an AI managing the resoure gathering operations.
    It is designed for the AIIDE06 tournamnent's game1, including:
    - Assuming there's only a single control center.
*/
class AutoGatherAI : public ClientModule
{
  //---------------------------------------------------------------------------
  /// constants
  //---------------------------------------------------------------------------
  //@{ 
private:
  static const sint4 FROM;                        ///< Identifier of this class when sending events

  static const std::string workerBPname;          ///< worker's blueprint name
  static const std::string mineralBPname;         ///< minerals' blueprint name
  static const std::string controlCenterBPname;   ///< control center's blueprint name

  /// Interval (ms) to wait before trying to reuse a mineral which cannot be reached
  static const uint4 MINERAL_SLEEP_TIME_MS;
  /// Maximum worker/mineral distance which allows mining
  static const sint4 WORKER_MINERAL_DISTANCE_THRESHOLD;
  /// Maximum worker/control-center distance which allows returning minerals
  static const sint4 WORKER_CC_DISTANCE_THRESHOLD;
  /// Maximum workers to associate with each mineral patch simultaneously
  static const uint4 MAX_WORKERS_PER_MINERAL;
  /// Minimal number of mineral clusters expected
  static const sint4 MIN_NUM_CLUSTERS;
  /// Maximal number of mineral clusters expected
  static const sint4 MAX_NUM_CLUSTERS;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Internal bookkeeping typedefs
  //---------------------------------------------------------------------------
  //@{
private:

  struct WorkerState {
    enum { IDLE, MOVING, MINING, RETURNINGMINERALS } job;
    const GobPtrWrapper* mineral;
    sint4 clusterId;
    sint4 clusterPathWaypoint;

    WorkerState() : job(IDLE), mineral(NULL), clusterId(-1), clusterPathWaypoint(0) {};
  };

  typedef std::set<GobPtrWrapper> GameObjects;
  typedef std::set<Worker>  Workers;
  typedef std::map<const Worker*,WorkerState> W2S;
  typedef std::multimap<const GobPtrWrapper*,Worker*> M2W;
  struct MineralCluster {
    Vector<const GobPtrWrapper*> mineralsInCluster;
    Vector<TerrainBase::Loc>     pathAroundCluster;
    /** For some clusters the path around might be blocked or out-of-bounds, 
	so use a fall-back to plain pathfinding. */
    bool usePathAroundCluster;
  };
  typedef Vector<MineralCluster> MineralClusters;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Variables
  //---------------------------------------------------------------------------
  //@{ 
private:
  Workers             workers;
  GameObjects         minerals;
  GameObjects         controlCenters;
  W2S                 workers2state;        ///< map a worker to its allotted mineral patch
  M2W                 minerals2workers;     ///< map a mineral patch to its mining workers
  MineralClusters     mineralClusters;
  uint4 minsMined;

  GameStateModule &gsm;  
  TerrainModule &tm;
  
  // boost::shared_ptr<SimpleTerrain::ST_Terrain> terrainimp; // declared before terrain for initialization order
  // boost::shared_ptr<TerrainModule> terrain;
  //@}
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  /// Internal bookkeeping
  //---------------------------------------------------------------------------
  //@{ 
  /// update world view with the given changes.
  void applyViewMessage(const GameChanges &changes);
  /// get the state of a given worker
  WorkerState& getState(const Worker& worker) { 
    W2S::iterator w2sIt = workers2state.find(&worker);
    assert(w2sIt != workers2state.end());
    return w2sIt->second;
  };
  const WorkerState& getState(const Worker& worker) const { 
    W2S::const_iterator w2sIt = workers2state.find(&worker);
    assert(w2sIt != workers2state.end());
    return w2sIt->second;
  };

  /// get the mineral associated with a given worker
  const GobPtrWrapper* getMineral(const Worker& worker) const { 
    return getState(worker).mineral;
  };

  /// remove given workers, minerals and control center(s) from the internal bookkeeping
  void removeWMC(const Vector<ScriptObj::WPtr>& objects);
  /// add given workers, minerals and control center(s) to the internal bookkeeping
  void insertWMC(const Vector<ScriptObj::WPtr>& objects);
  //@}
  //---------------------------------------------------------------------------

public:
  //----Game 3 Borrowing Functions----//
  void removeW(Vector<GameObj*>& objects);
  void insertW(Vector<GameObj*>& objects);
  uint4 getMinsMined(){uint4 temp; temp=minsMined; minsMined=0; return temp;}

private:
  //---------------------------------------------------------------------------
  /// Babysitting the workers
  //---------------------------------------------------------------------------
  //@{ 
  /// check if a worker is idle
  bool isIdle(const Worker& worker);
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Global planning
  //---------------------------------------------------------------------------
  //@{ 
private:
  /** Initialize the work plan. Should be called after the minerals and
      control center locations are known, ie after the first view msg was received. */
  void plan(void);
  /// Create a clusters map for the mineral locations
  void computeMineralClusters(void);
  /** Create roundabout paths for each mineral cluster, going to one side
      of the cluster and returning from the other side. 
      Assuming that the clusters were already computed. */
  void computePathsAroundClusters(void);

  /** choose and allocate a mineral patch to the given worker; 
      Optionally instructed to use a different patch than the one currently
      allocated, if at all possible. */
  void allocateMineralPatchFor(Worker& worker, bool forcedChange=false);
  /// associate a given worker and mineral in the internal maps
  void associateWorkerAndMineral(Worker& worker, const GobPtrWrapper& mineral);

  /// A comparison functor used for sorting mineral patches according to distance from a worker
  struct CompareMineralsByDistanceToWorker
  {
    const Worker& worker;
    CompareMineralsByDistanceToWorker(const Worker& worker_) : worker(worker_) {};
    bool operator()(const GobPtrWrapper* mineral1, const GobPtrWrapper* mineral2) {
      return Game::distance(*worker,**mineral1) < Game::distance(*worker,**mineral2);
    }
  };

  /** Check if a given mineral patch is assigned less than MAX_WORKERS_PER_MINERAL 
      workers, not counting a specific worker we're testing for. */
  bool isMineralFreeFor(const GobPtrWrapper* mineral, const Worker& worker) const;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Access a worker's status
  //---------------------------------------------------------------------------
  //@{ 
private:
  /// Is the given worker located near the control center? (No type-checking is done)
  bool isNearTheCC(const Worker& worker) const;
  /// Is the given worker located near its allotted mineral patch? (No type-checking is done)
  bool isNearTheMineral(const Worker& worker) const;
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Set a worker's action
  //---------------------------------------------------------------------------
  //@{ 
private:
  /// Move to the given location
  void moveTo(Worker& worker, const TerrainBase::Loc &location);
  /// Move to the given target
  void moveTo(Worker& worker, const GobPtrWrapper& target);
  /// Move back to the control center
  void moveToCC(Worker& worker);
  /// Move to the mineral patch allocated to this worker
  void moveToMineral(Worker& worker);
  /// Start mining (assuming the worker is near the mineral)
  void startMining(Worker& worker);
  /// Return resources (assuming the worker is near the control center)
  void returnMinerals(Worker& worker);
  //@}
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  /// Handlers for different events
  //---------------------------------------------------------------------------
  //@{
private:
  /// Handler for notifications of the PathFinder object
  class HandlerOfPathFinderEvents : public EventHandler {
    AutoGatherAI* parent;
  public:
    HandlerOfPathFinderEvents(AutoGatherAI* parent_) : parent(parent_) {};
    bool handle_event(const Event &event);
  } handlerOfPathFinderEvents;

  /// Handler for GameStateModule's view_msg notifications
  class HandlerOfViewMsgEvents : public EventHandler {
    AutoGatherAI* parent;
  public:
    HandlerOfViewMsgEvents(AutoGatherAI* parent_) : parent(parent_) {};
    bool handle_event(const Event &event);
  } handlerOfViewMsgEvents;
  //@}
  //---------------------------------------------------------------------------


private:
  /// Compute and apply actions to the workers
  void act();


  //---------------------------------------------------------------------------
  /// Public interface
  //---------------------------------------------------------------------------
  //@{
public:
  AutoGatherAI(GameStateModule &gsm_, TerrainModule &tm_);
  ~AutoGatherAI();
  //@}
  //---------------------------------------------------------------------------
};
#endif

