#ifndef IO_Buffer_H
#define IO_Buffer_H

// $Id: IO_Buffer.H,v 1.8 2006/04/27 14:00:56 ddeutscher Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// buffered socket I/O
// maintains message buffers that are sent/rcvd by threads that
// do blocking TCP I/O via SDL_net

#include "Global.H"

struct RandSrc;

class IO_Buffer {

public:

  enum { MAX_LEN = 1000000 }; // max. bytes per message uncompressed/compressed
  //  IO_Buffer() { rand_delay = 0; } // for the default call by derived classes

  IO_Buffer(TCPsocket so_, sint4 clev_, RandSrc *rand_delay_=0);
  virtual ~IO_Buffer();

  TCPsocket get_socket() const { return so; }

  // read/write status
  bool read_is_good() const { return read_good; }
  bool write_is_good() const { return write_good; }  

  // current buffer length
  sint4 write_buf_len();
  sint4 read_buf_len();
  
  // total number of bytes
  sint8 sent_bytes();
  sint8 rcvd_bytes();
  
  // signal threads to finish
  void finish() { done = 1; }

  // returns current size of o_buf (after adding the message)
  // or -1 if write error occurred  
  virtual sint4 write(const std::string &s);

  // returns current size of i_buf (before removing the message)
  // or -1 if read error occurred
  virtual sint4 read(std::string &s);

  // encode message (compress iff clev > 0)
  // return error code (0 iff OK)
  static sint4 encode(std::string &msg, sint4 clev);

  // return error code (0 iff OK)
  static sint4 decode(std::string &msg);

protected:
  
  TCPsocket so;
  sint4 clev;                          // compression level
  RandSrc *rand_delay;                 // for randomizing delays (=0: no delay)
  SDL_Thread *read_thread, *write_thread; // threads for blocking TCP I/O
  SDL_Thread *i_delay_thread;          // only used if delay_ms > 0

  SDL_mutex *buf_mut;           // protecting critical sections w.r.t. buf
  SDL_sem   *o_buf_sem;         // semaphore signaling msgs. ready to send
  SDL_mutex *i_delay_buf_mut;   // protecting critical sections w.r.t. i_delay_buf
  
  volatile bool done;                  // signals threads to finish
  volatile bool read_good, write_good; // true if reading/writing loop running
  
  std::list<std::string> i_buf, o_buf; // input/output message buffers
  uint1 *input_chars;                  // scratch space for socket reads
  volatile sint4 i_len, o_len;         // current string buffer length
  volatile sint8 i_bytes, o_bytes;     // bytes transferred

  sint4 blocking_write(void *buffer, sint4 len);
  sint4 blocking_read(void *buffer, sint4 len);  

  std::list<std::string> i_delay_buf, o_delay_buf;// buffers needed for delaying messages
  std::list<uint4> i_delay_times, o_delay_times;  // due times

public:

  static int write_loop(void *data);
  static int read_loop(void *data);  
  static int i_delay_loop(void *data);
};

#endif
