//Location.cpp

#include <cstdlib>
#include <math.h>
#include "Base.h"
#include <fstream>

//DJD: Point location function definitions {

//initializes each sector midpoint to point to the triangle in which it is contained
void TraDcdt::InitializeSectors()
{
	//get the bounds of the triangulation
	float xmax, xmin, ymax, ymin;
	get_bounds(xmin, xmax, ymin, ymax);
	//calculate the width and height of the triangulation
	float width = xmax - xmin;
	float height = ymax - ymin;
	//calculate the width and height of each sector
	sectorWidth = width / (float)xSectors;
	sectorHeight = height / (float)ySectors;
	//initialize all sectors to null
	for (int i = 0; i < ySectors; i++)
	{
		for (int j = 0; j < xSectors; j++)
		{
			sectors[i][j] = NULL;
		}
	}
	float x, y;
	float x1, y1, x2, y2, x3, y3;
	//go through the unprocessed triangles
	for (int k = 0; k < TraFace::Faces(); k++)
	{
		//retrieve the next valid triangle
		TraFace *currentFace = TraFace::Face(k);
		if (currentFace == NULL)
		{
			continue;
		}
		//get the minimum and maximum x and y values
		TriangleVertices(currentFace, x1, y1, x2, y2, x3, y3);
		float left = Min(x1, x2, x3);
		float right = Max(x1, x2, x3);
		float top = Min(y1, y2, y3);
		float bottom = Max(y1, y2, y3);
		//calculate the minimum and maximum x and y sector indices covered
		int xIndexMin = (int)((left - xmin + 0.5f * sectorWidth) / sectorWidth);
		int xIndexMax = (int)((right - xmin - 0.5f * sectorWidth) / sectorWidth);
		int yIndexMin = (int)((top - ymin + 0.5f * sectorHeight) / sectorHeight);
		int yIndexMax = (int)((bottom - ymin - 0.5f * sectorHeight) / sectorHeight);
		//go through the covered sector midpoints
		for (int i = yIndexMin; i <= yIndexMax; i++)
		{
			for (int j = xIndexMin; j <= xIndexMax; j++)
			{
				//calculate the location of the midpoint
				x = ((float)j + 0.5f) * sectorWidth + xmin;
				y = ((float)i + 0.5f) * sectorHeight + ymin;
				//if the midpoint is in the current triangle,
				if (InTriangle(currentFace, x, y))
				{
					//point the sector to that triangle
					sectors[i][j] = currentFace;
				}
			}
		}
	}
}

//check if the given point is within a given triangle
bool TraDcdt::InTriangle(TraFace *face, float x, float y)
{
	//get the triangle's vertices
	float x1, y1, x2, y2, x3, y3;
	TriangleVertices(face, x1, y1, x2, y2, x3, y3);
	//check that the point is counterclockwise of all edges
	return ((Orientation(x1, y1, x2, y2, x, y) >= 0)
		&& (Orientation(x2, y2, x3, y3, x, y) >= 0)
		&& (Orientation(x3, y3, x1, y1, x, y) >= 0));
}

//uses the sectors to locate in which triangle the given point lies
SeTriangulator::LocateResult TraDcdt::LocatePoint(float x, float y, TraBase* &result)
{
	//get the bounds of the triangulation
	float xmax, xmin, ymax, ymin;
	get_bounds(xmin, xmax, ymin, ymax);
	//calculates the indices of the sector in which the point lies
	int xIndex = (int)((x - xmin) / sectorWidth);
	int yIndex = (int)((y - ymin) / sectorHeight);
	//makes sure the indices are within the proper bounds
	xIndex = (xIndex < 0) ? 0 : (xIndex >= xSectors) ? xSectors - 1 : xIndex;
	yIndex = (yIndex < 0) ? 0 : (yIndex >= ySectors) ? ySectors - 1 : yIndex;
	//retrieve the triangle in which that sector midpoint is located
	SeFace *iniface = sectors[yIndex][xIndex];
	//if it was invalid, start at the usual place
	if (iniface == NULL)
	{
		iniface = _mesh->first()->fac();
	}
	//use this starting triangle to start point location
	return _triangulator->locate_point(iniface, x, y, result);
	//return LocatePoint((TraFace *)iniface, x, y, result);
}

//locate the given point starting at the usual place
SeTriangulator::LocateResult TraDcdt::LocatePointOld(float x, float y, TraBase* &result)
{
  return _triangulator->locate_point(_mesh->first()->fac(), x, y, result);
  //return LocatePoint((TraFace *)_mesh->first()->fac(), x, y, result);
}

//returns the maximum of the 3 values passed
float TraDcdt::Max(float a, float b, float c)
{
	return ((a >= b) && (a >= c)) ? a :
		((b >= a) && (b >= c)) ? b : c;
}

//returns the minimum of the 3 values passed
float TraDcdt::Min(float a, float b, float c)
{
	return ((a <= b) && (a <= c)) ? a :
		((b <= a) && (b <= c)) ? b : c;
}

//return a random value in the range of x values (used for testing)
float TraDcdt::RandomX()
{
	//get the bounds of the triangulation
	float xmax, xmin, ymax, ymin;
	get_bounds(xmin, xmax, ymin, ymax);
	return RandomBetween(xmin, xmax);
}

//return a random value in the range of y values (used for testing)
float TraDcdt::RandomY()
{
	//get the bounds of the triangulation
	float xmax, xmin, ymax, ymin;
	get_bounds(xmin, xmax, ymin, ymax);
	return RandomBetween(ymin, ymax);
}

//returns a random number between the values given
float TraDcdt::RandomBetween(float min, float max)
{
	float base = (float)rand() / (float)RAND_MAX;
	return (base * (max - min) + min);
}

//"walk" from the given triangle to the one specified
SeTriangulator::LocateResult TraDcdt::LocatePoint(TraFace *start, float x, float y, TraBase* &result)
{
	//get the bounds of the triangulation
	float xmax, xmin, ymax, ymin;
	get_bounds(xmin, xmax, ymin, ymax);
	//if the point is out of bounds, return that it could not be found
	if ((x < xmin) || (x > xmax) || (y < ymin) || (y > ymax))
	{
		result = NULL;
		return SeTriangulator::NotFound;
	}
	//keeps track of how many edges find the point counterclockwise
	//when this is 3, the point is inside the triangle
	int inside = 0;
	//get an edge originating from the vertex
	TraBase *s = start->se();
	//continues until it finds a triangle containing the point
	while (inside < 3)
	{
		//if the point is on the current vertex, return that
		if ((((TraVertex *)s->vtx())->p.x == x) && (((TraVertex *)s->vtx())->p.y == y))
		{
			result = s;
			return SeTriangulator::VertexFound;
		}
		//checks the orientation of the edge being examined with the point
		float orientation = Orientation(((TraVertex *)s->vtx())->p.x, ((TraVertex *)s->vtx())->p.y,
			((TraVertex *)s->nxt()->vtx())->p.x, ((TraVertex *)s->nxt()->vtx())->p.y, x, y);
		//if it's colinear,
		if (orientation == 0)
		{
			//check if the point is between the endpoints,
			//(ie. the point is closer to each endpoint than the endpoints are to each other)
			float xDiff0 = ((TraVertex *)s->vtx())->p.x - ((TraVertex *)s->nxt()->vtx())->p.x;
			float yDiff0 = ((TraVertex *)s->vtx())->p.y - ((TraVertex *)s->nxt()->vtx())->p.y;
			float xDiff1 = ((TraVertex *)s->vtx())->p.x - x;
			float yDiff1 = ((TraVertex *)s->vtx())->p.y - y;
			float xDiff2 = ((TraVertex *)s->nxt()->vtx())->p.x - x;
			float yDiff2 = ((TraVertex *)s->nxt()->vtx())->p.y - y;
			float Distance0 = xDiff0 * xDiff0 + yDiff0 * yDiff0;
			float Distance1 = xDiff1 * xDiff1 + yDiff1 * yDiff1;
			float Distance2 = xDiff2 * xDiff2 + yDiff2 * yDiff2;
			if ((Distance0 > Distance1) && (Distance0 > Distance2))
			{
				//if so, return that the point is on the edge
				result = s;
				return SeTriangulator::EdgeFound;
			}
		}
		//if it is counterclockwise from the edge to the point,
		else if (orientation >= 0)
		{
			//it is "inside" this edge
			inside++;
			//move to the next edge in the triangle
			s = s->nxt();
		}
		else
		{
			//otherwise, it wasn't in this triangle
			inside = 0;
			//move to the adjacent triangle and try again
			s = s->sym();
		}
	}
	//return the triangle containing the point given
	result = s;
	return SeTriangulator::TriangleFound;
}

//Point location function definitions }
