#ifndef Server_H
#define Server_H

// $Id: Server.H,v 1.19 2007/05/25 01:53:28 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// ORTS game server

#include "Global.H"
#include "TCP_Server.H"
#include "EventHandler.H"

class Options;
class Game;

// sliding window statistics

template <class T> class History {

public:

  Vector<T> h;
  sint4 max_len;
  T sum_;
  
  History(sint4 ml=8) { init(ml); }

  void init(sint4 ml) {
    max_len = ml;
    if (max_len < 1) ERR("max_len < 1");
    h.clear();
    sum_ = T(0);
  }
  
  void append(const T &x) {
    if (sint4(h.size()) >= max_len) {
      sum_ -= h.back();
      h.pop_back();
    }
    sum_ += x;
    h.insert(h.begin(), x);
  }

  T sum() const { return sum_; }

  T first() const {
    if (h.empty()) return T(0); else return h.back();
  }
  
  T prev() const {
    if (h.empty()) return T(0); else return h[0];
  }
  
  double avg() const {
    if (h.empty()) return T(0);
    // cout << "SUM=" << sum_ << " r=" << sum_/double(h.size()) << endl;
    return sum_/double(h.size());
  }

  sint4 length() const { return h.size(); }
  sint4 remaining() const { return max_len - h.size(); }
};


// worker mode login info

struct LoginInfo
{
  sint4 a,b,c,d; // ipaddr (high->low)
  std::string passwd;
};


class Server : public TCP_Server {

public:

  // see .C file for option comments
  
  struct Options {
    std::string host;
    sint4 port;
    sint4 clev;
    real8 freq;
    sint4 wait;   // <0:keyboard input, 0:full speed, >0:msec
    sint4 sdelay; // start delay in seconds
    sint4 tmax; 
    bool synchro;
    bool act;
    bool keepalive;
    sint4 seed; // game rng seed
    std::string load_replay;
    std::string save_replay;

    std::string save_file;
    std::string load_file;

    std::string ids; // client id:pw pairs
    std::string defines;

    Options();
    static void add();
  };
  
  
  // messages
  static const sint4 FROM;         // message origin
  static const sint4 READY_MSG;    // sent when game loop starts
  static const sint4 STEP_MSG;     // sent after views are sent
  

  enum { STAT_WINDOW = 8 };

  Server();
  ~Server();

  // add login info to list (used in worker mode, not checked yet)
  void add_login_info(const LoginInfo &li) { log_infos.push_back(li); }

  // wait for players to connect and run game
  void run(const std::string &world, Options &opt);

  // get access to the game
  Game &get_game() { return *game; }
  
  // event handlers
  void add_handler(EventHandler *h) { el.add(h); }
  void remove_handler(EventHandler *h) { el.remove(h); }
  int call_handlers(const Event &e) { return el.call(e); }

protected:

  Game *game;
  uint4 start_time;
  History<sint8> bytes_sent, bytes_rcvd, msgs_rcvd;
  History<sint8> frame_time;
  sint4 frame;
  uint1 save, load;
  FILE *save_file, *load_file;
  bool synchro; // true <=> server waits for messages
  Vector<LoginInfo> log_infos; // used in worker mode
  EventHandlerList el;
  
  // write status message
  void status(std::ostream &os) const;

  // send/recv messages, |msgs| = #players
  void send_messages(const Vector<std::string> &msgs);
  void recv_messages(Vector<std::string> &msgs);

  // messages from/to file
  void load_messages(Vector<std::string> &msgs);
  void save_messages(const Vector<std::string> &msgs);

  IO_Buffer *new_client(TCPsocket so, sint4 clev) const;

private:

  // forbid copy/assignment
  Server(const Server &) : TCP_Server() { ERR("what"); }
  Server &operator=(const Server &) { ERR("what"); }  
};

#endif
